from typing import Tuple
import numpy as np

from .misc.utils import isType


def svdAndReconstruction(A: np.ndarray, singularValues: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Generates an SVD decomposition of matrix A & reconstructs
    the matrix using these components and supplied singular values of matrix A

    :param A: matrix A of shape (m,m)
    :param singularValues(np.ndarray): wektor wartości singularnych (m,)


    :returns: (U, S, V, M), where U, S & V are SVD decomposition components & M is the reconstruction of A"""

    if not isType(A, np.ndarray) or not isType(singularValues, np.ndarray):
        raise ValueError("Wrong input types")

    m, shouldBeM = A.shape

    if m != shouldBeM or m != singularValues.shape[0]:
        raise ValueError("Wrong input shapes")

    U, S, V = np.linalg.svd(A)

    return U, S, V, (U * singularValues) @ V



from typing import Tuple, Union

import numpy as np

from .utils import isType


def exponential(x: Union[int, float], n: int) -> float:
    """
    Computes an approximation of e^x

    :param x: x
    :param k: k

    :return: the approximated value
    """

    acc = 0

    try:
        if isType(x, [int, float]) and isType(n, int) and n >= 0:
            for i in range(0, n):
                acc += pow(x, i) / np.math.factorial(i)

            return acc
        else:
            return np.nan
    except OverflowError:
        return acc


def coskx1(k: int, x: Union[int, float]) -> float:
    """
    Computes an approximation of cos(kx) using the formula:

    .. math::
        cos((m+1)x) = 2cos(x) * cos(mx) - cos((m-1)x)

    :param x: x
    :param k: k

    :return: the approximated value
    """

    try:
        if isType(k, int) and k >= 0 and isType(x, [int, float]):
            m1 = k - 1
            m2 = m1 - 1

            if m1 >= 1:
                if m1 == 1:
                    e1 = np.math.cos(x)
                else:
                    e1 = coskx1(m1, x)
            else:
                e1 = 1

            if m2 >= 1:
                if m2 == 1:
                    e2 = np.math.cos(x)
                else:
                    e2 = coskx1(m2, x)
            else:
                e2 = 1

            if k == 0:
                return np.math.cos(0)
            elif k == 1:
                return np.math.cos(x)
            else:
                return 2 * np.math.cos(x) * e1 - e2
        else:
            return np.nan
    except:
        return np.nan


def cossinkx2(k: int, x: Union[int, float]) -> Tuple[float, float]:
    """
    Computes an approximation of cos(kx) & sin(kx) using the formulas:

    .. math::
        cos(mx) = cosx \\cdot cos(m-1)x - sinx \\cdot sin(m-1)x
    .. math::
        sin(mx) = sinx \\cdot cos(m-1)x + cosx \\cdot sin(m-1)x

    :param x: x
    :param k: k

    :return: approximated cos(kx) and sin(kx) values
    """

    try:
        if isType(k, int) and k >= 0 and isType(x, [int, float]):
            if k == 0:
                return (np.math.cos(0), np.math.sin(0))
            elif k == 1:
                return (np.math.cos(x), np.math.sin(x))
            else:
                (cosKLess1, sinKLess1) = cossinkx2((k - 1), x)
                return (np.math.cos(x) * cosKLess1 - np.math.sin(x) * sinKLess1, np.math.sin(x) * cosKLess1 + np.math.cos(x) * sinKLess1)
        else:
            return np.nan, np.nan
    except:
        return np.nan, np.nan


def pi(n: int) -> float:
    """Computes an approximate value of Pi, using the formula:

    .. math::
        \\sum_{n=1}^{\\infty} \\cfrac{1}{n^2} = \\cfrac{1}{6} \\pi^2


    :param n: the amount of sum elements

    :return: the approximated value of pi
    """

    if isType(n, int) and n > 0:
        acc = 0

        for i in range(1, n + 1):
            acc += 1 / pow(i, 2)

        return np.math.sqrt(acc * 6)
    else:
        return np.nan



from setuptools import setup, find_packages

# Load long_description from README.md
with open("README.md") as f:
    long_description = f.read()

setup(
    name='mnultitool',
    version='1.0.4',
    description='A multitool for Numerical Methods',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/artus9033/MNultitool',
    author='artus9033',
    author_email='artus9033@gmail.com',
    classifiers=[
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='numerical,math,interpolation',
    packages=find_packages(),
    python_requires='>=3.6, <4',
)



from typing import Any, Tuple, cast
from matplotlib.axes import Axes
import numpy as np
import matplotlib.pyplot as plt


def plotVector(title: str, ax, vec: np.ndarray, showLegendColorbar: bool = True) -> None:
    """
    Plots a vector colorbar in a transposed form, with an additional legend of values, if needed

    :param title: the title of the figure (or axes, if on a subplot)
    :param ax: Axes object
    :param vec: the vector
    :param showLegendColorbar: whether to show the legend colorbar, optional
    """

    ax.set_title(title)
    cax = ax.matshow(np.transpose(vec.reshape(len(vec), 1)))

    if showLegendColorbar:
        ax.colorbar(cax, location="bottom")


def plotMatrixAndVector(A: np.ndarray, b: np.ndarray, bigTitle: str = "Visualization of matrix A & vector b", matrixTitle: str = "Matrix A", vectorTitle: str = "Vector b", figsize=None) -> Tuple[Any, Tuple[Axes, Axes]]:
    """
    Plots a colormap presenting matrix A and a colorbar presenting vector b, with a legend of the range of values

    :param A: the matrix to be presented
    :param b: the vector to be presented; _note: the vector will be copied & transposed for visualization_
    :param bigTitle: the suptitle of the figure, optional
    :param matrixTitle: the title of the subplot of the matrix, optional
    :param vectorTitle: the title of the subplot of the vector, optional
    :param figsize: the size of the plot, optional

    :return: a tuple containing the figure and a tuple of the two axes
    """

    if not figsize:
        figsize = (12, 10)

    fig, (ax1, ax2) = cast(Any, plt.subplots(
        2, 1, constrained_layout=True, figsize=(figsize)))

    fig.suptitle(bigTitle)

    ax1.set_title(matrixTitle)
    cax1 = ax1.matshow(A)
    fig.colorbar(cax1, location="top")

    fig.show()

    return fig, (ax1, ax2)
